import { Address } from "ton-core";
import { Log } from "../../../logs/Log.js";
import { OpCode } from "../../OpCode.js";
import { UnknownWallet } from "../../wallets/UnknownWallet.js";
import { DedustBuyAction } from "../actions/DedustBuyAction.js";
import { DedustSellAction } from "../actions/DedustSellAction.js";
import { JettonTransferAction } from "../actions/JettonTransferAction.js";
import { JettonTransferNotificationAction } from "../actions/JettonTransferNotificationAction.js";
import { Message } from "../Message.js";
import { Transaction } from "../Transaction.js";
export class TonwebJettonTransactionFetcher {
    account;
    blockchain;
    // wtf???
    supportedOpCodes = [
        OpCode.jetton_transfer_notification,
        OpCode.jetton_transfer,
        OpCode.dedust_buy,
        OpCode.dedust_sell,
    ];
    ignoredOpCodes = [
        OpCode.dedust_unknown,
        OpCode.dedust_unknown2,
    ];
    constructor(account, blockchain) {
        this.account = account;
        this.blockchain = blockchain;
        for (const [index, opCode] of this.supportedOpCodes.entries()) {
            this.supportedOpCodes[index] = new blockchain.client_tonweb.utils.BN(opCode);
        }
    }
    async fetchTransactions(limit) {
        const transactions = await this.fetch(limit);
        return this.parse(transactions);
    }
    isOpSupported(op) {
        for (const opCode of this.supportedOpCodes) {
            if (op.eq(opCode)) {
                return true;
            }
        }
        return false;
    }
    isOpIgnoring(op) {
        for (const opCode of this.ignoredOpCodes) {
            if (op.toString() === opCode.toString()) {
                return true;
            }
        }
        return false;
    }
    parseJettonTransferNotificationAction(slice, tonweb) {
        const queryId = slice.loadUint(64);
        const amount = slice.loadCoins();
        const from = slice.loadAddress();
        const maybeRef = slice.loadBit();
        const payload = maybeRef ? slice.loadRef() : slice;
        const payloadOp = payload.loadUint(32);
        if (!payloadOp.eq(new tonweb.utils.BN(0))) {
            Log.info("no text comment in transfer_notification");
        }
        const payloadBytes = payload.loadBits(slice.getFreeBits());
        const comment = new TextDecoder().decode(payloadBytes);
        const value = BigInt(amount.toString());
        const sender = new UnknownWallet(Address.parse(from.toString()));
        return new JettonTransferNotificationAction(value, sender);
    }
    parseJettonTransferAction(slice, tonweb) {
        const queryId = slice.loadUint(64);
        const amount = slice.loadCoins();
        const destination = slice.loadAddress();
        const respDestination = slice.loadAddress();
        /* const maybeCustomPayload = slice.loadBit();
            const customPayload = maybeCustomPayload ? slice.loadRef() : slice;*/
        // const forwardTonAmount = slice.loadUint(16);
        const maybePayload = slice.loadBit();
        const payload = maybePayload ? slice.loadRef() : slice;
        const payloadOp = payload.loadUint(32);
        if (!payloadOp.eq(new tonweb.utils.BN(0))) {
            Log.info("no text comment in transfer");
        }
        const payloadBytes = payload.loadBits(slice.getFreeBits());
        const comment = new TextDecoder().decode(payloadBytes);
        const value = BigInt(amount.toString());
        const dest = new UnknownWallet(Address.parse(destination.toString()));
        const respDest = new UnknownWallet(Address.parse(respDestination.toString()));
        return new JettonTransferAction(value, dest, respDest);
    }
    parseAction(msgBody, tonweb) {
        const cell = tonweb.boc.Cell.oneFromBoc(msgBody);
        const slice = cell.beginParse();
        const op = slice.loadUint(32);
        if (!this.isOpSupported(op)) {
            if (!this.isOpIgnoring(op)) {
                Log.warn("Op code " +
                    op.toString() +
                    " not supported! Hex calculator: https://defuse.ca/big-number-calculator.htm");
            }
            return;
        }
        if (op.eq(new tonweb.utils.BN(OpCode.jetton_transfer_notification))) {
            Log.info("Jetton transfer notification found!");
            return this.parseJettonTransferNotificationAction(slice, tonweb);
        }
        else if (op.eq(new tonweb.utils.BN(OpCode.jetton_transfer))) {
            Log.info("Jetton transfer found!");
            return this.parseJettonTransferAction(slice, tonweb);
        }
        else if (op.eq(new tonweb.utils.BN(OpCode.dedust_buy))) {
            Log.info("DeDust buy found!");
            return this.parseDedustBuyAction();
        }
        else if (op.eq(new tonweb.utils.BN(OpCode.dedust_sell))) {
            Log.info("Dedust sell found!");
            return this.parseDedustSellAction();
        }
        else {
            Log.error("Called not supported op code " + op.toString());
        }
    }
    parseDedustBuyAction() {
        return new DedustBuyAction();
    }
    parseDedustSellAction() {
        return new DedustSellAction();
    }
    parseMessage(msg, tonweb) {
        const sourceAddress = msg.source;
        const destinationAddress = msg.destination;
        if (!msg.msg_data ||
            msg.msg_data["@type"] !== "msg.dataRaw" ||
            !msg.msg_data.body) {
            // no msg or msg body
            Log.info("Message not found!");
            return;
        }
        const source = new UnknownWallet(sourceAddress);
        const destination = new UnknownWallet(destinationAddress);
        const value = BigInt(msg.value);
        const bounced = false; // TODO: load bounced https://docs.ton.org/develop/data-formats/msg-tlb
        const msgBody = tonweb.utils.base64ToBytes(msg.msg_data.body);
        const action = this.parseAction(msgBody, tonweb);
        return new Message(source, destination, value, bounced, action);
    }
    parse(transactions) {
        const tonweb = this.blockchain.client_tonweb;
        // tslint:disable-next-line:new-parens
        const result = new Array();
        for (const tx of transactions) {
            // tslint:disable-next-line:radix
            const time = new Date(parseInt(tx.utime) * 1000);
            const fee = tx.fee;
            const hash = tx.transaction_id.hash;
            const lt = tx.transaction_id.lt;
            // tslint:disable-next-line:variable-name
            const in_msg = this.parseMessage(tx.in_msg, tonweb);
            // tslint:disable-next-line:variable-name new-parens
            const out_msgs = new Array();
            for (const msg of tx.out_msgs) {
                // tslint:disable-next-line:variable-name
                const out_msg = this.parseMessage(msg, tonweb);
                if (out_msg === undefined) {
                    Log.warn("Skipping out_msgs of " + hash);
                    continue;
                }
                out_msgs.push(out_msg);
            }
            const transaction = new Transaction(time, fee, hash, lt, in_msg, out_msgs);
            result.push(transaction);
        }
        return result;
    }
    async fetch(limit) {
        const tonweb = this.blockchain.client_tonweb;
        return await tonweb.provider.getTransactions(this.account.address.toString(), limit);
    }
}
